@@ -1,5 +1,5 @@
 /* This file is part of GNU cflow
-   Copyright (C) 1997,2005,2007 Sergey Poznyakoff
+   Copyright (C) 1997-2021 Sergey Poznyakoff
  
    GNU cflow is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,10 +11,8 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
-   You should have received a copy of the GNU General Public
-   License along with GNU cflow; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
-   MA 02110-1301 USA */
+   You should have received a copy of the GNU General Public License
+   along with this program.  If not, see <http://www.gnu.org/licenses/>. */
 
 #ifdef HAVE_CONFIG_H
 # include <config.h>
@@ -43,17 +41,28 @@
 # define setlocale(category, locale) /* empty */
 #endif
 
+/* Exit codes */
+#define EX_OK    0  /* Success */
+#define EX_FATAL 1  /* Fatal error */
+#define EX_SOFT  2  /* Some input files cannot be read or parsed */
+#define EX_USAGE 3  /* Command line usage error */
+
 #define NUMITEMS(a) sizeof(a)/sizeof((a)[0])
 
-typedef struct cons *Consptr;
-typedef struct cons Cons;
-struct cons {
-     Consptr car;
-     Consptr cdr;
+struct linked_list_entry {
+     struct linked_list_entry *next, *prev;
+     struct linked_list *list;
+     void *data;
 };
 
-#define CAR(a) (a)->car
-#define CDR(a) (a)->cdr
+typedef void (*linked_list_free_data_fp) (void*);
+
+struct linked_list {
+     linked_list_free_data_fp free_data;
+     struct linked_list_entry *head, *tail;
+};
+
+#define linked_list_head(list) ((list) ? (list)->head : NULL)
 
 enum symtype {
      SymUndefined,  /* Undefined or deleted symbol */
@@ -76,19 +85,27 @@ typedef struct {
 
 enum symbol_flag {
      symbol_none,
-     symbol_temp,                  /* Temporary symbol. Must be deleted after
-				      processing of the current module */
-     symbol_parm                   /* Parameter */
+     symbol_local,                 /* Unit-local symbol. Must be deleted after
+				      processing current compilation unit */
+     symbol_parm,                  /* Parameter */
+     symbol_alias                  /* Alias to another symbol */
 };
 
 typedef struct symbol Symbol;
 
 struct symbol {
+     struct table_entry *owner;
      Symbol *next;                 /* Next symbol with the same hash */
+     struct linked_list_entry *entry;
+     
      enum symtype type;            /* Type of the symbol */
      char *name;                   /* Identifier */
      enum symbol_flag flag;        /* Specific flag */
-     
+     struct symbol *alias;         /* Points to the aliased symbol if
+				      type==SymToken and flag==symbol_alias.
+				      In this case, the rest of the structure
+				      is ignored */
+	
      int active;                   /* Set to 1 when the symbol's subtree is
 				      being processed, prevent recursion */
      int expand_line;              /* Output line when this symbol was first
@@ -97,7 +114,7 @@ struct symbol {
      int token_type;               /* Type of the token */
      char *source;                 /* Source file */
      int def_line;                 /* Source line */
-     Consptr ref_line;             /* Referenced in */
+     struct linked_list *ref_line; /* Referenced in */
      
      int level;                    /* Block nesting level (for local vars),
 				      Parameter nesting level (for params) */
@@ -109,8 +126,9 @@ struct symbol {
 				      variables */  
      
      int recursive;                /* Is the function recursive */
-     Consptr caller;               /* List of callers */
-     Consptr callee;               /* List of callees */
+     size_t ord;                   /* ordinal number */
+     struct linked_list *caller;   /* List of callers */
+     struct linked_list *callee;   /* List of callees */
 };
 
 /* Output flags */
@@ -143,6 +161,7 @@ extern int brief_listing;
 extern int reverse_tree;
 extern int out_line;
 extern char *start_name;
+extern int all_functions;
 extern int max_depth;
 extern int emacs_option;
 extern int debug;
@@ -156,16 +175,32 @@ extern int token_stack_increase;
 extern int symbol_count;
 extern unsigned input_file_count;
 
-Symbol *lookup(char*);
-Symbol *install(char*);
+#define INSTALL_DEFAULT     0x00
+#define INSTALL_OVERWRITE   0x01
+#define INSTALL_CHECK_LOCAL 0x02
+#define INSTALL_UNIT_LOCAL  0x04
+
+Symbol *lookup(const char*);
+Symbol *install(char*, int);
+Symbol *install_ident(char *name, enum storage storage);
+void ident_change_storage(Symbol *sp, enum storage storage);
 void delete_autos(int level);
 void delete_statics(void);
 void delete_parms(int level);
 void move_parms(int level);
-void cleanup(void);
-int collect_symbols(Symbol ***, int (*sel)());
-Consptr append_to_list(Consptr *, void *);
-int symbol_in_list(Symbol *sym, Consptr list);
+size_t collect_symbols(Symbol ***, int (*sel)(), size_t rescnt);
+size_t collect_functions(Symbol ***return_sym);
+struct linked_list *linked_list_create(linked_list_free_data_fp fun);
+void linked_list_destroy(struct linked_list **plist);
+void linked_list_append(struct linked_list **plist, void *data);
+void linked_list_prepend(struct linked_list **plist, void *data);
+void linked_list_iterate(struct linked_list **plist, 
+			 int (*itr) (void *, void *), void *data);
+void linked_list_unlink(struct linked_list *list,
+			struct linked_list_entry *ent);
+size_t linked_list_size(struct linked_list *list);
+
+int data_in_list(void *data, struct linked_list *list);
 
 int get_token(void);
 int source(char *name);
@@ -175,12 +210,14 @@ void pp_option(const char *arg);
 
 void init_parse(void);
 int yyparse(void);
+void reset_static_caller(void);
 
 void output(void);
 void newline(void);
 void print_level(int lev, int last);
 int globals_only(void);
 int include_symbol(Symbol *sym);
+int symbol_is_function(Symbol *sym);
 
 void sourcerc(int *, char ***);
 
@@ -216,3 +253,9 @@ int posix_output_handler(cflow_output_command cmd,
 			 FILE *outfile, int line,
 			 void *data, void *handler_data);
 
+
+typedef struct cflow_depmap *cflow_depmap_t;
+cflow_depmap_t depmap_alloc(size_t count);
+void depmap_set(cflow_depmap_t dmap, size_t row, size_t col);
+int depmap_isset(cflow_depmap_t dmap, size_t row, size_t col);
+void depmap_tc(cflow_depmap_t dmap);
