@@ -1,5 +1,5 @@
 /* This file is part of GNU cflow
-   Copyright (C) 1997, 2005, 2006, 2007, 2009, 2010, 2011 Sergey Poznyakoff
+   Copyright (C) 1997-2021 Sergey Poznyakoff
  
    GNU cflow is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,13 +11,12 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
-   You should have received a copy of the GNU General Public
-   License along with GNU cflow; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
-   MA 02110-1301 USA */
+   You should have received a copy of the GNU General Public License
+   along with this program.  If not, see <http://www.gnu.org/licenses/>. */
 
 #include <cflow.h>
 #include <parser.h>
+#include <ctype.h>
 
 typedef struct {
      char *name;
@@ -48,11 +47,11 @@ void maybe_parm_list(int *parm_cnt_return);
 void call(char*, int);
 void reference(char*, int);
 
-int level;                  /* Current nesting level */
-Symbol *caller;             /* Current caller */
-struct obstack text_stk;    /* Obstack for composing declaration line */
+static int level;                  /* Current nesting level */
+static Symbol *caller;             /* Current caller */
+static struct obstack text_stk;    /* Obstack for composing declaration line */
 
-int parm_level;             /* Parameter declaration nesting level */
+static int parm_level;             /* Parameter declaration nesting level */
 
 typedef struct {
      int type;
@@ -62,10 +61,10 @@ typedef struct {
 
 typedef int Stackpos[1];
 
-TOKSTK tok;
-TOKSTK *token_stack;
-int tos;
-int curs;
+static TOKSTK tok;
+static TOKSTK *token_stack;
+static int tos;
+static int curs;
 int token_stack_length = 64;
 int token_stack_increase = 32;
 static int need_space;
@@ -114,6 +113,94 @@ print_token(TOKSTK *tokptr)
      }
 }
 
+static char *
+token_type_str(int t)
+{
+     static char buf[80];
+     switch (t) {
+     case 0:
+	  return "EOF";
+     case WORD:
+	  return "WORD";
+     case LBRACE0:
+	  return "'{'";
+     case RBRACE0:
+	  return "'}'";
+     case IDENTIFIER:
+	  return "IDENTIFIER";
+     case EXTERN:
+	  return "EXTERN";
+     case STATIC:
+	  return "STATIC";
+     case TYPEDEF:
+	  return "TYPEDEF";
+     case STRUCT:
+	  return "STRUCT";
+     case MODIFIER:
+	  return "MODIFIER";
+     case OP:
+	  return "OP";
+     case UNION:
+	  return "UNION";
+     case ENUM:
+	  return "ENUM";
+     case LBRACE:
+	  return "' {'";
+     case RBRACE:
+	  return "' }'";
+     case MEMBER_OF:
+	  return "MEMBER_OF";
+     case TYPE:
+	  return "TYPE";
+     case STRING:
+	  return "STRING";
+     case PARM_WRAPPER:
+	  return "PARM_WRAPPER";
+     case QUALIFIER:
+	  return "QUALIFIER";
+     }	  
+     if (isprint(t))
+	  snprintf(buf, sizeof(buf), "'%c'(%d)", t, t);
+     else
+	  snprintf(buf, sizeof(buf), "%d", t);
+     return buf;
+}
+
+static void
+dbgtok(TOKSTK *t, int delim)
+{
+     if (delim)
+	  fputc(delim, stderr);
+     fprintf(stderr, "{ %s ", token_type_str(t->type));
+     if (t->type)
+	  fprintf(stderr, ", %s, %d ", t->token ? t->token : "NULL", t->line);
+     fputc('}', stderr);
+}
+
+static void
+debugtoken(TOKSTK *t, char *fmt, ...)
+{
+     if (debug > 1) {
+	  va_list ap;
+	  int i;
+	  
+	  if (fmt) {
+	       va_start(ap, fmt);
+	       vfprintf(stderr, fmt, ap);
+	       va_end(ap);
+	       fprintf(stderr, ": ");
+	  }
+	  if (t) {
+	       dbgtok(t, 0);
+	       fprintf(stderr, "; ");
+	  }
+	  fprintf(stderr, "%d: {", curs);
+	  for (i = curs; i < tos; i++)
+	       dbgtok(token_stack + i, i == curs ? 0 : ',');
+	  fprintf(stderr, "}\n");
+     }
+}
+
 static void
 file_error(char *msg, TOKSTK *tokptr)
 {
@@ -129,14 +216,45 @@ void
 mark(Stackpos pos)
 {
      pos[0] = curs;
+     if (debug > 1)
+	  fprintf(stderr, "marking stack at %d\n", curs);
 }
 
 void
 restore(Stackpos pos)
 {
      curs = pos[0];
-     if (curs)
+     if (curs) {
 	  tok = token_stack[curs-1];
+	  debugtoken(&tok, "restored stack");
+     }
+}
+
+void
+tokdel(int beg, int end)
+{
+     if (end >= beg) {
+	  if (end < tos)
+	       memmove(token_stack + beg, token_stack + end + 1,
+		       (tos - end - 1) * sizeof(token_stack[0]));
+	  tos -= (end - beg + 1);
+     }
+}
+
+void
+tokins(int pos, int type, int line, char *token)
+{
+     if (++tos == token_stack_length) {
+	  token_stack_length += token_stack_increase;
+	  token_stack = xrealloc(token_stack,
+				 token_stack_length*sizeof(*token_stack));
+     }
+     memmove(token_stack + pos + 1, token_stack + pos,
+	     (tos - pos - 1) * sizeof(token_stack[0]));
+     token_stack[pos].type = type;
+     token_stack[pos].token = token;
+     token_stack[pos].line = line;
+     debugtoken(&token_stack[pos], "insert at %d", pos);
 }
 
 void
@@ -157,9 +275,10 @@ cleanup_stack()
 {
      int delta = tos - curs;
 
-     if (delta) 
+     if (delta > 0) 
 	  memmove(token_stack, token_stack+curs, delta*sizeof(token_stack[0]));
-
+     else if (delta < 0) /* Invalid input */
+	  delta = 0;
      tos = delta;
      curs = 0;
 }
@@ -178,23 +297,27 @@ nexttoken()
      if (curs == tos) {
 	  type = get_token();
 	  tokpush(type, line_num, yylval.str);
+	  yylval.str = NULL;
      }
      tok = token_stack[curs];
      curs++;
+     debugtoken(&tok, "next token");
      return tok.type;
 }
 
 int
 putback()
 {
-     if (curs == 0)
-	  error(10, 0, _("INTERNAL ERROR: cannot return token to stream"));
+     if (curs == 0) {
+	  error(0, 0, _("INTERNAL ERROR: cannot return token to stream"));
+	  abort();
+     }
      curs--;
      if (curs > 0) {
-	  tok.type = token_stack[curs-1].type;
-	  tok.token = token_stack[curs-1].token;
+	  tok = token_stack[curs-1];
      } else
 	  tok.type = 0;
+     debugtoken(&tok, "putback");
      return tok.type;
 }
 
@@ -256,6 +379,20 @@ save_token(TOKSTK *tokptr)
 	  obstack_1grow(&text_stk, tokptr->type);
 	  need_space = 0;
 	  break;
+     case LBRACE:
+     case LBRACE0:
+	  if (need_space) 
+	       obstack_1grow(&text_stk, ' ');
+	  obstack_1grow(&text_stk, '{');
+	  need_space = 1;
+	  break;
+     case RBRACE:
+     case RBRACE0:
+	  if (need_space) 
+	       obstack_1grow(&text_stk, ' ');
+	  obstack_1grow(&text_stk, '}');
+	  need_space = 1;
+	  break;
      case OP:
 	  obstack_1grow(&text_stk, ' ');
 	  obstack_grow(&text_stk, tokptr->token, strlen(tokptr->token));
@@ -283,6 +420,12 @@ undo_save_stack()
      save_end = -1;
 }
 
+int
+save_stack_is_empty()
+{
+     return save_end <= 0;
+}
+	  
 char *
 finish_save_stack(char *name)
 {
@@ -405,6 +548,9 @@ is_function()
 	  case MODIFIER:
 	  case STATIC:
 	  case EXTERN:
+	  case STRUCT:
+	  case UNION:
+	  case ENUM:
 	       nexttoken();
 	       continue;
 	  case PARM_WRAPPER:
@@ -473,9 +619,17 @@ expression()
 		    parens_lev++;
 	       } else {
 		    reference(name, line);
+
+		    /* MEMBER_OF can be preceded by a closing paren, as in
+		       (*a)->b
+		    */
+		    while (parens_lev > 0 && tok.type == ')') {
+			 parens_lev--;
+			 nexttoken();
+		    }
+
 		    if (tok.type == MEMBER_OF) {
-			 while (tok.type == MEMBER_OF)
-			      nexttoken();
+			 nexttoken();
 		    } else {
 			 putback();
 		    }
@@ -483,7 +637,7 @@ expression()
 	       break;
 	  case '(':
 	       /* maybe typecast */
-	       if (nexttoken() == TYPE)
+	       if (nexttoken() == TYPE || tok.type == STRUCT)
 		    skip_to(')');
 	       else {
 		    putback();
@@ -493,6 +647,9 @@ expression()
 	  case ')':
 	       parens_lev--;
 	       break;
+	  case MEMBER_OF:
+	       nexttoken();
+	       break;
 	  }
 	  nexttoken();
      }
@@ -528,6 +685,8 @@ parse_function_declaration(Ident *ident, int parm)
      case LBRACE:
 	  if (ident->name) {
 	       caller = lookup(ident->name);
+	       if (caller && caller->storage == AutoStorage)
+		    caller = NULL;
 	       func_body();
 	  }
 	  break;
@@ -540,29 +699,15 @@ parse_function_declaration(Ident *ident, int parm)
 int
 fake_struct(Ident *ident)
 {
-     Stackpos sp;
-     
-     mark(sp);
      ident->type_end = -1;
      if (tok.type == STRUCT) {
 	  if (nexttoken() == IDENTIFIER) {
-	       ident->type_end = tos;
+	       ident->type_end = curs;
 	  }
 	  putback();
 	  skip_struct();
-	  if (tok.type == IDENTIFIER || tok.type == MODIFIER) {
-	       TOKSTK hold = tok;
-	       restore(sp);
-	       if (ident->type_end == -1) {
-		    /* there was no tag. Insert { ... } */
-		    tos = curs;
-		    token_stack[curs].type = IDENTIFIER;
-		    token_stack[curs].token = "{ ... }";
-		    tos++;
-	       } else {
-		    tos = curs + 1;
-	       }
-	       tokpush(hold.type, hold.line, hold.token);
+	  if (tok.type == IDENTIFIER || tok.type == MODIFIER || tok.type == QUALIFIER) {
+	       putback();
 	  } else if (tok.type == '(')
 	       return 0;
 	  else if (tok.type != ';')
@@ -579,27 +724,23 @@ parse_variable_declaration(Ident *ident, int parm)
      
      mark(sp);
      ident->type_end = -1;
-     if (tok.type == STRUCT) {
+     if (tok.type == STRUCT || tok.type == UNION) {
 	  if (nexttoken() == IDENTIFIER) {
-	       ident->type_end = tos;
+	       ident->type_end = curs;
 	  }
 	  putback();
 	  skip_struct();
 	  while (tok.type == MODIFIER || tok.type == QUALIFIER)
 	       nexttoken();
 	  if (tok.type == IDENTIFIER) {
-	       TOKSTK hold = tok;
-	       restore(sp);
 	       if (ident->type_end == -1) {
+		    int pos = curs-1;
+		    restore(sp);
 		    /* there was no tag. Insert { ... } */
-		    tos = curs;
-		    token_stack[curs].type = IDENTIFIER;
-		    token_stack[curs].token = "{ ... }";
-		    tos++;
-	       } else {
-		    tos = curs + 1;
+		    tokdel(curs, pos - 1);
+		    tokins(curs, IDENTIFIER, tok.line, "{ ... }");
+		    debugtoken(&tok, "modified stack");
 	       }
-	       tokpush(hold.type, hold.line, hold.token);
 	  } else {
 	       if (tok.type == ';')
 		    return;
@@ -635,7 +776,6 @@ parse_variable_declaration(Ident *ident, int parm)
 	  else
 	       expression();
 	  goto select;
-	  break;
      case LBRACE0:
      case LBRACE:
 	  func_body();
@@ -700,8 +840,10 @@ skip_struct()
      }
 
      while (tok.type == PARM_WRAPPER) {
-	  if (skip_balanced('(', ')', 0) == -1)
+	  if (skip_balanced('(', ')', 0) == -1) {
 	      file_error(_("unexpected end of file in struct"), NULL);
+	      return;
+	  }
      }
 }
 
@@ -960,6 +1102,7 @@ get_knr_args(Ident *ident)
      Ident id;
 
      switch (tok.type) {
+     case QUALIFIER:
      case IDENTIFIER:
      case TYPE:
      case STRUCT:
@@ -978,6 +1121,7 @@ get_knr_args(Ident *ident)
 		    stop = 1;
 		    break;
 	       case TYPE:
+	       case QUALIFIER:
 	       case IDENTIFIER:
 	       case STRUCT:
 		    putback();
@@ -1026,7 +1170,8 @@ declare(Ident *ident, int maybe_knr)
 	  && (!maybe_knr || get_knr_args(ident) == 0)
 	  && !(tok.type == LBRACE || tok.type == LBRACE0 || tok.type == TYPE
 	       || tok.type == PARM_WRAPPER))
-	 || (ident->parmcnt < 0 && ident->storage == ExplicitExternStorage)) {
+	 || (ident->parmcnt < 0 && ident->storage == ExplicitExternStorage)
+	 || save_stack_is_empty()) {
 	  undo_save_stack();
 	  /* add_external()?? */
 	  return;
@@ -1061,7 +1206,7 @@ declare(Ident *ident, int maybe_knr)
      sp->def_line = ident->line;
      sp->level = level;
      if (debug)
-	  printf(_("%s:%d: %s/%d defined to %s\n"),
+	  fprintf(stderr, _("%s:%d: %s/%d defined to %s\n"),
 		 filename,
 		 line_num,
 		 ident->name, ident->parmcnt,
@@ -1086,7 +1231,8 @@ declare_type(Ident *ident)
      sp->def_line = ident->line;
      sp->ref_line = NULL;
      if (debug)
-	  printf(_("%s:%d: type %s\n"), filename, line_num, ident->name);
+	  fprintf(stderr, _("%s:%d: type %s\n"), filename, line_num,
+		  ident->name);
 }
 
 Symbol *
@@ -1156,3 +1302,14 @@ reference(char *name, int line)
      }
 }
 
+/* If the caller is a static symbol, reset it to NULL.  This function
+   is called as a part of cleanup routine at the end of compilation
+   module. */
+void
+reset_static_caller(void)
+{
+     if (caller &&
+	 (caller->storage == StaticStorage || caller->flag == symbol_local))
+	  caller = NULL;
+}
+
